package de.mutschler.rpdl;

import de.mutschler.util.*;
import java_cup.runtime.*;
import de.mutschler.rpdl.basicvalue.BasicValue;
import de.mutschler.rpdl.fileimport.*;
import de.mutschler.rpdl.entity.*;
import de.mutschler.rpdl.command.RpdlCommand;
import de.mutschler.rpdl.calc.*;
import de.mutschler.rpdl.tree.*;
import java.io.*;
import java.util.*;

/**
 * This is the basic class, which handles a complete
 * instance of the RPDL
 */
public class RpdlBase extends LogObj
{
	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------

	//----------------------------------------------------------------------
	//------- Attribute Definitions
	//----------------------------------------------------------------------
	/** The container of the attibutedefinitions. The mapping is(String)name -> (RpdlAttributeDef)attributedefinition */
	private Map attibuteDef = new HashMap();
	
	/**
	 * Get an attribute definition.
	 * @param name the name to identify the attribute
	 * @return the attribute, or null, if it isn't found
	 */
	public RpdlAttributeDef getAttributeDef(String name) {
		return (RpdlAttributeDef)attibuteDef.get(name);
	}

	/**
	 * Add an attribute definition.
	 * @param name the name to identify the attribute
	 * @return the previous attribute, or null, if it didn't exist yet
	 * @exception RpdlException if there already exists an attribute definition
	 *                          of the same name.
	 */
	public RpdlAttributeDef addAttributeDef(RpdlAttributeDef ad) throws RpdlException {
		// check, if the type exists.
		BasicValue b = BasicValue.createValue(ad.getType());
		// check, if the definition already exists
		if (attibuteDef.containsKey(ad.getName()))
			throw new RpdlException("attribute definition '" + ad.getName() + "' already exists!");
		return (RpdlAttributeDef)attibuteDef.put(ad.getName(), ad);
	}

	/** Just print some debug infos. */
	public void displayAttrDef() {
		StringBuffer sb = new StringBuffer(2000);
		sb.append(CR + "******************************************************" + CR);
		int i = 1;
		for (Iterator it = attibuteDef.values().iterator(); it.hasNext(); ) {
			RpdlAttributeDef ad = (RpdlAttributeDef)it.next();
			sb.append("AttrDef #" + i + ": " );
			sb.append( ad.getType() + " " + ad.getName());
			sb.append(CR);
			i++;
		}
		warn(sb.toString());
	}
	//----------------------------------------------------------------------
	//------- Base Object
	//----------------------------------------------------------------------

	/** The basetype. */
	private RpdlType baseType = null;
	/** The base-object */
	private RpdlObject baseObject = null;	// in the beginning this is not defined


	/** Get the base-object.
	 * @return the base object, or null, if it hasn't been set yet.
	 */
	public RpdlObject getBaseObject() {return baseObject;}
	
	/** Set the base Object. Note: if the basetype is not set yet,
	 * or the object is not the right type, then an exception is thrown.
	 */
	public void setBaseObject(RpdlObject base) throws RpdlException {
		if (baseType == null)
			throw new RpdlException("Basetype not set yet");
		if (!base.isType(baseType))
			throw new RpdlException("Baseobject type mismatch: object '" + base.getName() +
									"' is not of type " + baseType.getName());
		if (baseObject != null)
			throw new RpdlException("Base already set with object '" + baseObject.getName() + "'." +
									" Object '" + base.getName() + "' must not be of type '" + getBaseType().getName() + "'.");
		debug("Using object " + base.getName() + " as the base-object!");
		baseObject = base;
		base.setBaseObject(base);
	}

	/** Get the base type. If not set, null is returned. */
	public RpdlType getBaseType() {return baseType;}
	
	/** Set the baseType. This can only be done once! */
	public void setBaseType(RpdlType type) throws RpdlException {
		if (baseType != null) throw new RpdlException("Basetype already set!");
		baseType = type;
	}
	
	/** Just print some debug infos. */
	public void displayBase() {
		StringBuffer sb = new StringBuffer(200);
		sb.append(CR + "******************************************************" + CR);
		sb.append("Basetype = " + getBaseType().getName() + CR);
		sb.append("Base-Object Tree:" + CR);
		if (getBaseObject() != null)
			sb.append(subObjectsString(getBaseObject()));
		else
			sb.append("   no base object defined");
		sb.append(CR);
		warn(sb.toString());
	}

	/**
	 * put the objects in the form
	 * + sub1
	 * | + sub11
	 * | + sub12
	 * + sub2
	 *   + sub21
	 */
	private String subObjectsString(RpdlObject obj) {
		TreeFormatter tf = new TreeFormatter() {
			public String getInfo(Object o)		{return ((RpdlObject)o).getName();}
			public Iterator subitems(Object o)	{return ((RpdlObject)o).getSubObjects();}
		};
		return tf.getTreeString(obj);
	}
	
	//----------------------------------------------------------------------
	//------- Tables
	//----------------------------------------------------------------------

	/** The Tables. The mapping is (String)name -> (RpdlTable)table*/
	private Map tables = new HashMap();
	
	/** Get a table.
	 * @return the table, or null, if it doesn't exists
	 */
	public RpdlTable getTable(String name) {return (RpdlTable)tables.get(name);}

	/** Add a table.
	 * @return the previous table, or null, in none exists before
	 */
	public RpdlTable addTable(RpdlTable tab) throws RpdlException {
		// check, if the table already exists
		if (tables.containsKey(tab.getName()))
			throw new RpdlException("Table '" + tab.getName() + "' already exists!");
		return (RpdlTable)tables.put(tab.getName(), tab);
	}

	/** Just print a list of tables. */
	public void displayTables() {
		StringBuffer sb = new StringBuffer(2000);
		sb.append(CR + "******************************************************" + CR);
		int i = 1;
		for (Iterator it = tables.values().iterator(); it.hasNext(); ) {
			RpdlTable t = (RpdlTable)it.next();
			sb.append("Table #" + i + ": " );
			sb.append(t.getName());
			sb.append(CR);
			i++;
		}
		warn(sb.toString());
	}

	//----------------------------------------------------------------------
	//------- History
	//----------------------------------------------------------------------

	/** The storage of the history. This is an ordered list, which contains RpdlHistoryEntry-objects. */
	private List history = new LinkedList();
	
	/** Get the HistoryEntries in chronological order. */
	public Iterator getHistoryEntries() {return history.iterator();}
	
	/** Add an entry to the history. */
	public void addHistoryEntry(RpdlHistoryEntry he) {history.add(he);}
	
	/** Get the latest history entry. This is used for associating
	 *  a command with a history.
	 */
	public RpdlHistoryEntry getCurrentHistoryEntry()
	{return (RpdlHistoryEntry)history.get(history.size()-1);}
	
	/** Just print a list of the history. */
	public void displayHistory() {
		StringBuffer sb = new StringBuffer(2000);
		sb.append(CR + "******************************************************" + CR);
		int i = 1;
		for (Iterator it = getHistoryEntries(); it.hasNext(); ) {
			RpdlHistoryEntry h = (RpdlHistoryEntry)it.next();
			sb.append("History #" + i + ": " );
			sb.append(h.toString());
			sb.append(CR);
			i++;
		}
		warn(sb.toString());
	}

	//----------------------------------------------------------------------
	//------- Object Library
	//----------------------------------------------------------------------

	/** Container to store all the objects of the character. */
	private Map library = new HashMap();
	
	/** Get an Object. If the object does not exist, null is returned. */
	public RpdlObject getLibObject(String name) {return (RpdlObject)library.get(name);}

	/** Add an Object. An exception is thrown, if the object already exists. */
	public void addLibObject(RpdlObject ro) throws RpdlException
	{
		String name = ro.getName();
		if (library.containsKey(name)) {
			throw new RpdlException("Object '" + name + "' already exists!");
		}
		library.put(name, ro);
	}
		
	/** Just print a list of objects. */
	public void displayLibObjects() {
		StringBuffer sb = new StringBuffer(2000);
		sb.append(CR + "******************************************************" + CR);
		int i = 1;
		for (Iterator it = library.values().iterator(); it.hasNext(); ) {
			RpdlObject o = (RpdlObject)it.next();
			sb.append("LibObject #" + i + ": " + "  -------------------------------------" + CR);
			sb.append(o.toString());
			sb.append(CR);
			i++;
		}
		warn(sb.toString());
	}
	//----------------------------------------------------------------------
	//------- Objects (Character)
	//----------------------------------------------------------------------

	/** Container to store all the objects of the character. */
	private Map objects = new HashMap();
	
	/** Get an Object. If the object doesn't exist, null is returned. */
	public RpdlObject getObject(String name) {return (RpdlObject)objects.get(name);}

	/** add an Object. An Exception is thrown if the object already exists. */
	public void addObject(RpdlObject ro) throws RpdlException
	{
		String name = ro.getName();
		if (objects.containsKey(name)) {
			throw new RpdlException("Object '" + name + "' already exists!");
		}
		objects.put(name, ro);
		
		// check if the object is of the base-type
		RpdlType basetype = getBaseType();
		if (basetype != null && ro.isType(baseType)) {
			setBaseObject(ro);
		}
		triggerObjectChange(RpdlBase.NOTIFY_OBJECT_CREATE, ro);
	}
		
	/** Just print a list of objects. */
	public void displayObjects() {
		StringBuffer sb = new StringBuffer(4000);
		sb.append(CR + "******************************************************" + CR);
		int i = 1;
		for (Iterator it = objects.values().iterator(); it.hasNext(); ) {
			RpdlObject o = (RpdlObject)it.next();
			sb.append("Object #" + i + ": " + "  -------------------------------------" + CR);
			sb.append(o.toString());
			sb.append(CR);
			i++;
		}
		warn(sb.toString());
	}

	//--------------- The trigger list for object changes
	/**
	 * List of StmtHandles to trigger, when an Object changes.
	 * Items in this list are of type <code>StmtHandle</code>.
	 * The events can be: created/modified/deleted
	 */
	private List notifyObjectList = new LinkedList();
	
	/** formatted triggerList. */
//	String triggerListString() {return listString(triggerStatements());}

	/** Get all the statements, which will be triggered by this attribute. */
	private Iterator notifyObjectStatements() {return notifyObjectList.iterator();}
	
	/** Add a StmtHandle to the triggerObjectList */
	public void addNotifyObject(StmtHandle sh) throws RpdlException {
		if (notifyObjectList.contains(sh))
			throw new RpdlException("Handle " + sh + " already in notifyObjectList!");
		ddebug("added " + sh+ " to notifyObjectList");
		notifyObjectList.add(sh);
	}

	/** Object notification for creation. */
	public final static int NOTIFY_OBJECT_CREATE=0;
	/** Object notification for change. */
	public final static int NOTIFY_OBJECT_CHANGE=1;
	/** Object notification for deletion. */
	public final static int NOTIFY_OBJECT_DELETE=2;

	/** Get a string for the notification type. useful for debugging. */
	public static String getNotifyObjectTypeName(int type) {
		switch(type) {
			case NOTIFY_OBJECT_CREATE: return "CREATE";
			case NOTIFY_OBJECT_CHANGE: return "CHANGE";
			case NOTIFY_OBJECT_DELETE: return "DELETE";
			default: return "UNKNOWN";
		}
	}

	/** fire the notification events */
	public void triggerObjectChange(int type, RpdlObject ro) {
		debug("Triggering object " + getNotifyObjectTypeName(type).toLowerCase() + " for Object '" + ro.getName() +"'");
		for (Iterator it = notifyObjectStatements(); it.hasNext(); ) {
			StmtHandle sh = (StmtHandle)it.next();
			try {
				sh.trigger(this);
			} catch (RpdlException re) {
				addError(re);
			}
		}
	}
	
	//----------------------------------------------------------------------
	//------- Types
	//----------------------------------------------------------------------

	/** The types. The Mapping is (String)typename -> (RpdlType)type */
	private Map types = new HashMap();
	
	/** Get a Type. If the type doesn't exit, null is returned. */
	public RpdlType getType(String name) {return (RpdlType)types.get(name);}

	/**
	 * Add a Type.
	 * This methods checks:
	 * <ul><li>if the type already exists
	 * <li> if the parenttypes exist
	 * </ul>
	 */
	public void addType(RpdlType rt) throws RpdlException
	{
		String name = rt.getName();
		if (types.containsKey(name)) {
			throw new RpdlException("Type '" + name + "' already exists!");
		}
		types.put(name, rt);
	}

	/** just print a list of types*/
	public void displayTypes() {
		StringBuffer sb = new StringBuffer(200);
		sb.append(CR + "******************************************************" + CR);
		int i = 1;
		for (Iterator it = types.values().iterator(); it.hasNext(); ) {
			RpdlType t = (RpdlType)it.next();
			sb.append("Type #" + i + ": "  + "  -------------------------------------" + CR);
			sb.append(t.toString());
			sb.append(CR);
			i++;
		}
		warn(sb.toString());
	}
	
	//----------------------------------------------------------------------
	//------- Event handling
	//----------------------------------------------------------------------

	/*
	 * When a statements makes link in the character, all the necessary upadtes are
	 * queued in the event list. When the addCommand is finished, then the event
	 * list is parsed, and all the dependecies are recomputed.
	 */
	
	private CalcHandler calcHandler = new CalcHandler(this);
	
	/**
	 * Queue a new CalcItem for calculation.
	 * @see CalcHandler#addCalcItem
	 */
	public final void addCalcItem(CalcItem ci) throws RpdlException
	{calcHandler.addCalcItem(ci);}

	/**
	 * Remove a queued caclulation. This might be the case, if an
	 * attribute is removed, and therefore can't be calculated anymore.
	 */
	public void removeCalcItem(CalcItem ci) {calcHandler.removeCalcItem(ci);}

	public final int getNumIgnored() {return calcHandler.ignoreCnt;}
	public final int getNumCalc() {return calcHandler.calcCnt;}
	public final int getCalcTime() {return calcHandler.calcTime;}

	//----------------------------------------------------------------------
	//------- ObjectOrder: The order of the items
	//----------------------------------------------------------------------

	/** The current ordering. */
	private ObjectOrder oOrder;
	{setObjectOrder(new FlatOrder());}	// by default use the FlatScanner
	
	/** Set a new TreeWalker. This performs a recalculation of everything! */
	public void setObjectOrder(ObjectOrder oo) {
		oOrder = oo;
		oo.setMain(this);
	}
	
	/** Get the TreeWalker. This object defines the ordering of
	 *  Objects and Statements.*/
	public ObjectOrder getObjectOrder() {return oOrder;}

	//----------------------------------------------------------------------
	//------- Error handling
	//----------------------------------------------------------------------
	
	private List errorList = new LinkedList();
	
	/** clear the error list. */
	public void clearErrorList() {errorList.clear();}
	
	/** Check if there were errors. */
	public boolean hasError() {return !errorList.isEmpty();}
	
	/** add an error. */
	public void addError(Exception e) {
		addError(e.getMessage());
	}
	
	/** add an error. */
	public void addError(String err) {
		if (!errorList.contains(err))	// prevent duplicate entries
			errorList.add(err);
	}

	/** Get the errors. */
	public Iterator getErrors() {return errorList.iterator();}
	
	
	//----------------------------------------------------------------------
	//------- Base functionality
	//----------------------------------------------------------------------
	
	/**
	 * Suspend the execution. If you reenable this, all queued
	 * calculations are performed. So, suspend(false) will always execute
	 * the current calculations
	 * @param suspend: if true, execution is suspended, if false exeution is reenabled.
	 * @see CalcHandler#suspend
	 */
	public void suspend(boolean suspend) throws RpdlException {
		calcHandler.suspend(suspend);
	}

	/**
	 * This is the main method, to add a command to the current RPDL-object.
	 * @see de.mutschler.rpdl.command
	 */
	public void addCommand(RpdlCommand cmd) {
		if (cmd == null) {
			warn("ignoring null command.");
			return;
		}
		try {
			cmd.add(this);
			debug("added command " + cmd);
			calcHandler.execute();
		} catch (RpdlException re) {
			error("while adding command: " + re.getMessage());
			addError(re);
			if (Log.getLogLevel() >= Log.DEBUG)
				Log.exception(re);
		}
	}

	/** For convinience: it simply calls RpdlConfig.getConfig(). */
	public static final RpdlConfig getConfig() {return RpdlConfig.getConfig();}
	
	//----------------------------------------------------------------------
	//------- Misc
	//----------------------------------------------------------------------
	
	/** Display all collections of this class. This method simply
	 * calls all display*-Methods
	 */
	public void displayAll() {
		displayHistory();
		displayAttrDef();
		displayTables();
		displayBase();
		displayLibObjects();
		displayObjects();
		displayTypes();
	}
	
}