package de.mutschler.rpdl.basicvalue;

import de.mutschler.util.*;
import de.mutschler.rpdl.RpdlException;
import java.util.HashMap;

/**
 * For the math-functions, every subclass has to implement a logic
 * if the operations between the types can be done, and the result
 * is a new instance of the BasicValue. 
 */
public abstract class BasicValue extends LogObj implements Cloneable,java.io.Serializable
{
	static final long serialVersionUID = 4316406712114619483L;
	//----------------------------------------------------------------------
	//------- Constants
	//----------------------------------------------------------------------
	/** add two numbers */
	public static final int OP_ADD = 0;
	/** subtract two numbers */
	public static final int OP_SUB = 1;
	/** multiply two numbers */
	public static final int OP_MUL = 2;
	/** divide two numbers */
	public static final int OP_DIV = 3;
	/** Negate the number */
	public static final int OP_NEG = 4;
	/** comparison '=' */
	public static final int COMP_EQ= 5;
	/** comparison '<>' */
	public static final int COMP_NE= 6;
	/** comparison '<' */
	public static final int COMP_LT= 7;
	/** comparison '<=' */
	public static final int COMP_LE= 8;
	/** comparison '>' */
	public static final int COMP_GT= 9;
	/** comparison '>=' */
	public static final int COMP_GE=10;
	/** conversion to STRING */
	public static final int TO_STRING=11;
	/** conversion to BOOL */
	public static final int TO_BOOL = 12;
	/** conversion to INT */
	public static final int TO_INT =  13;
	/** conversion to NUMBER */
	public static final int TO_NUMBER=14;
	/** conversion to DICE */
	public static final int TO_DICE = 15;
	

	/** The name of this type. */
	public static final String TYPE_ERR="ERROR";

	/** store here a list of avail types. The mapping is string -> BasicValue. */
	private static HashMap availValues = new HashMap();
	
	static {
		Log.debug("de.mutschler.rpdl.basicvalue.BasicValue", "Initializing types");
		BasicValue t;
		t = new ValueBool();	availValues.put(t.getType(), t);
		t = new ValueDice();	availValues.put(t.getType(), t);
		t = new ValueInt();		availValues.put(t.getType(), t);
		t = new ValueNumber();	availValues.put(t.getType(), t);
		t = new ValueString();	availValues.put(t.getType(), t);
	}
	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------
	public BasicValue() {super();}
	public BasicValue(String s) {super();}

	//----------------------------------------------------------------------
	//------- Accessors
	//----------------------------------------------------------------------

	/** Get a name of this type. This is used to identify the type.
	 * usually this is a fixed string.
	 */
	public abstract String getType();
	
	/**
	 * Set this value to the value of the other object.
	 */
	public void setValue(BasicValue t) throws RpdlConversionException
	{throw new RpdlConversionException("assignment from type '" + t.getType() + "' to type '" + getType() + "' not supported");}

	/** Get a string-representation of the value. This is used for concatenation. */
	public abstract String getStringValue();

	//----------------------------------------------------------------------
	//------- mathematical operations
	//----------------------------------------------------------------------

	/** Perform a calculation. */
	public BasicValue calculate(int op, BasicValue p2) throws RpdlConversionException {
		if (op == TO_STRING)	// TO_STRING is always implemented
			return new ValueString(getStringValue());
		throw new RpdlConversionException("Operation " + getType() + " " + getOpName(op) + (p2==null?"":" " + p2.getType()) + " not supported!");
	}
	
	/**
	 * For the static type-checking. Return the expected type for an
	 * operation.
	 */
	public String getResultType(int op, String type) {
		if (op == TO_STRING)	// TO_STRING is always implemented
			return ValueString.TYPENAME;
		return TYPE_ERR;
	}
	
	/** Create a new instance of a BasicValue, with a given type. */
	public static BasicValue createValue(String type) throws RpdlException {
		BasicValue t = (BasicValue)availValues.get(type);
		if (t == null) throw new RpdlException("Failed creating value of type " + type);
		return (BasicValue)t.clone();
	}
	/**
	 * Get the result-type for this kind of operation.
	 */
	public static String getResultType(int op, String type1, String type2)
	{
		BasicValue t1 = (BasicValue)availValues.get(type1);
		if (t1 == null) return TYPE_ERR;
		return t1.getResultType(op, type2);
	}
	
	/** Get the operation to cast into the specific type. */
	public static int getCastOperation(String type) {
		BasicValue v = (BasicValue)availValues.get(type);
		return v.getCastOp();
	}
	
	/** Return the cast operation for casting into the same value type. */
	public abstract int getCastOp();
	
	public abstract Object clone();

	/** Turn the constant into a useful string for displaying */
	public static String getOpName(int op) {
		String s;
		switch(op) {
			case OP_ADD:  s = " +"; break;
			case OP_SUB:  s = " -"; break;
			case OP_MUL:  s = " *"; break;
			case OP_DIV:  s = " /"; break;
			case OP_NEG:  s = "-"; break;
			case COMP_EQ: s = "=="; break;
			case COMP_NE: s = "!="; break;
			case COMP_LT: s = "< "; break;
			case COMP_LE: s = "<="; break;
			case COMP_GT: s = "> "; break;
			case COMP_GE: s = ">="; break;
			case TO_STRING: s = "(S)"; break;
			case TO_BOOL:	s = "(B)"; break;
			case TO_INT:	s = "(I)"; break;
			case TO_NUMBER: s = "(N)"; break;
			case TO_DICE:	s = "(D)"; break;
			default:    	s = "?"+op+"?"; break;
		}
		return s;
	}
	
	// default for toString()
	public String toString() {return getStringValue();}
}
