package de.mutschler.rpdl.entity;

import de.mutschler.rpdl.*;
import de.mutschler.rpdl.entity.*;
import de.mutschler.rpdl.statement.Stmt;
import de.mutschler.rpdl.tree.ObjectOrder;

import de.mutschler.util.LogObj;

/**
 * This is a class, which binds a statement to an RpdlObject.
 * It can be used as a listener.
 */

public class StmtHandle extends LogObj implements Comparable
{
	//----------------------------------------------------------------------
	//------- Attributes
	//----------------------------------------------------------------------
	
	private RpdlObject object;
	private Stmt statement;
	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------

	public StmtHandle(RpdlObject obj, Stmt stmt) {
		object = obj;
		statement = stmt;
	}
	
	/** Get the associated object. */
	public final RpdlObject getObject() {return object;}
	/** Get the associated statement. */
	public final Stmt getStatement() {return statement;}

	/**
	 * Execute a statement.
	 * This called is passed to Stmt.execute().
	 * @see Stmt#execute
	 */
	public void execute(RpdlBase main) throws RpdlException {
		getStatement().execute(main, getObject());
	}

	/**
	 * This method is called, when an object is linked/unlinked to the base object.
	 * This called is passed to Stmt.onBaseLink().
	 * @param link true if linked, false if unlinked
	 * @see Stmt#onBaseLink
	 */
	public void onBaseLink(RpdlBase main, boolean link) throws RpdlException {
		getStatement().onBaseLink(main, getObject(), this, link);
	}
	
	/**
	 * Trigger this statement for recalculation.
	 * This called is passed to Stmt.trigger().
	 * @see Stmt#trigger
	 */	
	public void trigger(RpdlBase main) throws RpdlException {
		getStatement().trigger(main, this);
	}

	/**
	 * Compare the Stmthandle against another.
	 * If the current Stmt should be executed before the given one,
	 * a negative number is returned.
	 */
	public int compareTo(Object o) {
		StmtHandle sh2 = ((StmtHandle)o);
		Stmt st1 = getStatement();
		Stmt st2 = sh2.getStatement();
		if (st1.getPriority() < st2.getPriority()) return -1;
		if (st1.getPriority() > st2.getPriority()) return  1;
//		debug("stmt1 '" + this + "' has same pri as '" + sh2 +"'");
		// priority is equal. Now compare the objects
		RpdlObject ro1 = getObject();
		RpdlObject ro2 = sh2.getObject();
		ObjectOrder oo = ro1.getMain().getObjectOrder();
		int res = oo.compare(ro1, ro2);
		if (res != 0) return res;
		// hmm. same object, same priority
		// now test the IDs
		int id1 = st1.getID();
		int id2 = st2.getID();
		return (id1<id2)?-1:(id1==id2?0:1);
	}
	
	
	/**
	 * Two handles are the same, if they have the same object,
	 * and the same statement.
	 */
	public boolean equals(Object o) {
		StmtHandle sh = (StmtHandle) o;
		return	getStatement().equals(sh.getStatement()) &&
				getObject().equals(sh.getObject());
	}
	
	/** for debugging. */
	public String toString() {
		return "StmtHandle obj=" + getObject().getName() + ", stmt=" + getStatement();
	}

}