package de.mutschler.rpdl.tree;

import de.mutschler.rpdl.*;
import de.mutschler.rpdl.entity.RpdlObject;
import de.mutschler.rpdl.entity.StmtHandle;
import de.mutschler.util.LogObj;
import java.util.*;

/**
 * This is the abstract class which contains the functions
 * for traversing the object-tree in different orders.
 * Therefore the two main funktions are handled:
 * <ol>
 * <li>Traversing the tree</li>
 * <li>comparing two elements of the tree, and making a total order.</li>
 * </ol>
 */
public abstract class ObjectOrder extends LogObj implements Comparator
{
	/** Set the RpdlBase, this Treewalker operates on. */
	public void setMain(RpdlBase main) {this.main = main;}
	
	private RpdlBase main;
	
	/** Get the RpdlBase. */
	protected final RpdlBase getMain() { return main;}
	
	
	/**
	 * This method is for traversing. An Iterator is returned,
	 * which can be used for looking at each element.
	 * by default, this gets all objects from the Base-object,
	 * and performs a sorting on it. There are no duplicate
	 * objects.
	 */
	public Iterator walk() {
		// This code is untested, since I never used it....
		// get all objects in the base
		SortedSet objects = new TreeSet(this);
		addObject(objects, getMain().getBaseObject());
		return objects.iterator();
	}
	
	/** Recursively add this object, all all decendents. */
	private void addObject(SortedSet ss, RpdlObject ro) {
		if (ro == null) return;
		ss.add(ro);
		for (Iterator it = ro.getSubObjects(); it.hasNext();) {
			addObject(ss, (RpdlObject)it.next());
		}
	}
	
	/**
	 * This method should be called, to notify the treewalker of changes
	 * in the tree. Here you can update internal data.
	 */
	public void treeChanged() {}
	
	//---- Comparator interface
	
	/**
	 * This method is for comparing two rpdl-objects in a tree.
	 * The method returns a negative int, if the first object
	 * is less than the second object.
	 * <p>
	 * This method handles objects of class RpdlObject. The
	 * corresponding compare methods are called then.
	 * @see #compare(RpdlObject, RpdlObject)
	 */
	public int compare(Object o1, Object o2) {
		// compare objects
		if (o1 instanceof RpdlObject) {
			return compare((RpdlObject)o1, (RpdlObject)o2);
		}
		// error
		throw new IllegalArgumentException("Cannot compare object of class " + o1.getClass().getName());
	}
	
	/**
	 * This method is for comparing two rpdl-objects in a tree.
	 * The method returns a negative int, if the first object
	 * is less than the second object.
	 */
	public abstract int compare (RpdlObject ro1, RpdlObject ro2);

	/**
	 * Check if we have the same comparator. Not really useful in this case.
	 */
	public boolean equals(Object o1) {
		return o1 == this;
	}
	
}