package de.mutschler.util;

import java.util.Iterator;
import de.mutschler.util.StringUtil;

/**
 * This method formats a tree in the following format:
 * + sub1
 * | + sub11
 * | + sub12
 * + sub2
 *   + sub21
 *
 * To use this class, implement the getInfo() and subitems() methods,
 * and then you can call getTreeString(), to get the formatted tree as
 * as a String.
 *
 * <pre>
 * Example (from Entity):
 *	public String parentTypesString() {
 * 		TreeFormatter tf = new TreeFormatter() {
 * 			public String getInfo(Object o)		{return ((Entity)o).getName();}
 *			public Iterator subitems(Object o)	{return ((Entity)o).parentTypes();}
 *		};
 *		return tf.getTreeString(this);
 *	}
 * </pre>
 */

public abstract class TreeFormatter
{
	//---- Methods to be implemented
	/** Get the info to be displayed from teh object. */
	public abstract String getInfo(Object o);
	
	/** Get subitems of the same type from the object. */
	public abstract Iterator subitems(Object o);
	

	/** This is the same as getSubTreeString, but it displays
	 * the root item as well.
	 * <pre>
	 * root
	 * + sub1
	 * | + sub11
	 * | + sub12
	 * + sub2
	 *   + sub21
	 * </pre>
	 */
	public String getTreeString(Object root) {
		return getInfo(root) + LogObj.CR + getSubTreeString(root);
	}
	
	/**
	 * create a tree of the form
	 * <pre>
	 * + sub1
	 * | + sub11
	 * | + sub12
	 * + sub2
	 *   + sub21
	 * </pre>
	 */
	public String getSubTreeString(Object root) {
		StringBuffer sb = new StringBuffer(100);
		String sub=null;
		boolean start = true;
		for (Iterator it = subitems(root); it.hasNext(); ) {
			Object o = it.next();
			if (!StringUtil.isEmpty(sub)) {
				sb.append(LogObj.CR);
				sb.append(StringUtil.indent(sub, "| "));
			}
			if (!start) sb.append(LogObj.CR);
			start = false;
			// take care of newlines in the string...
			sb.append(StringUtil.indent(getInfo(o), "+ ", "  "));
			sub = getSubTreeString(o);
		}
		if (!StringUtil.isEmpty(sub)) {
			sb.append(LogObj.CR);
			sb.append(StringUtil.indent(sub, "  "));
		}
		return sb.toString();
	}
}